# Copyright 2015 Jeff Vierstra

from .bisect import *

import numpy as np
import scipy.stats


def emperical_fdr(pvals_null, pvals):
	"""
	"""

	''' DEPRECATED
	sorted_pvals_null = np.sort(pvals_null, axis = 0)
	sorted_pvals_idx = np.argsort(pvals)

	n = np.arange(1, pvals.shape[0]+1, dtype = np.float64)

	counts = np.apply_along_axis(lambda x: bisect(x, np.asarray(pvals)[sorted_pvals_idx]), 0, sorted_pvals_null)
	false_positive_rates = np.mean(counts, axis = 1) / n
	
	false_positive_rates[false_positive_rates > 1] = 1

	return false_positive_rates[np.argsort(sorted_pvals_idx)]
	'''

	sorted_pvals_null = np.sort(np.ravel(pvals_null))
	sorted_pvals_idx = np.argsort(pvals)

	counts = bisect(sorted_pvals_null, pvals[sorted_pvals_idx])
	false_positive_rates = counts / len(sorted_pvals_null)
	false_positive_rates[false_positive_rates > 1] = 1
	return false_positive_rates[np.argsort(sorted_pvals_idx)]

# Not sure if Storey FDR is best (appears to be too conservative)

def pi0est(pvals, lamb = None):
	""" """

	lamb = np.arange(0.05, 1, 0.05) if lamb is None else lamb
	
	n = len(pvals)

	pi0 = np.array([np.mean(pvals>=l)/(1-l) for l in lamb])

	#bootstrap method
	minpi0 = np.percentile(pi0, q = 10)
	W = np.array([np.sum(pvals >= l) for l in lamb])
	mse = (W / (n**2 * (1-lamb)**2)) * (1 - W/n) + (pi0 - minpi0)**2

	return min(pi0[mse== np.min(mse)], 1)
	#return lamb

def qvalue(pvals):
	""" """

	pi0 = pi0est(pvals)

	n = len(pvals)
	u = np.argsort(pvals)
	v = scipy.stats.rankdata(pvals, method = 'max')
	#qvals = (pi0*n*pvals) / v
	qvals = (pi0*n*pvals) / (v * (1 - (1-pvals)**n))
	qvals[u[n-1]] = min(qvals[u[n-1]], 1)
	for i in range(0, n-1)[::-1]:
		qvals[u[i]] = min(qvals[u[i]], qvals[u[i+1]])

	return qvals

def bh_qvalue(pvals):
    """
    Return Benjamini-Hochberg FDR q-values corresponding to p-values C{pv}.

    This function implements an algorithm equivalent to L{bh_rejected} but
    yields a list of 'adjusted p-values', allowing for rejection decisions
    based on any given threshold.

    @type pv: list
    @param pv: p-values from a multiple statistical test

    @rtype: list
    @return: adjusted p-values to be compared directly with the desired FDR
      level
    """
    m = len(pvals)
    rank, sorted_pvals = list(zip(*sorted(enumerate(pvals), None, operator.itemgetter(1))))
    if pvals[0] < 0 or pvals[-1] > 1:
        raise ValueError("p-values must be between 0 and 1")
    qvalues = np.zeros(m)
    mincoeff = sorted_pvals[-1]
    qvalues[rank[-1]] = mincoeff
    for j in range(m-2, -1, -1):
        coeff = m*sorted_pvals[j]/float(j+1)
        if coeff < mincoeff:
            mincoeff = coeff
        qvalues[rank[j]] = mincoeff
    return qvalues




# Copyright 2015 Jeff Vierstra

# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: embedsignature=True

from footprint_tools.stats.distributions import nbinom
import numpy as np
from scipy import optimize

import pwlf

cimport cython
cimport numpy as np

from footprint_tools.stats.distributions cimport nbinom


cpdef data_type_t piecewise_three(x, data_type_t x0, data_type_t x1, data_type_t x2,
	 								data_type_t y0, data_type_t y1, data_type_t y2,
	 								data_type_t k0, data_type_t k1, data_type_t k2):
	
	return (
        (x<x0) * (y0 + k0*(x)) +
        ((x>=x0) & (x<x1)) * (y1 + k1*(x)) +
        (x>=x1) * (y2 + k2*(x))
    )

cpdef data_type_t piecewise_four(x, data_type_t x0, data_type_t x1, data_type_t x2, data_type_t x3, 
	 								data_type_t y0, data_type_t y1, data_type_t y2, data_type_t y3,
	 								data_type_t k0, data_type_t k1, data_type_t k2, data_type_t k3):
	
	return (
        (x<x0) * (y0 + k0*(x)) +
        ((x>=x0) & (x<x1)) * (y1 + k1*(x)) +
        ((x>=x1) & (x<x2)) * (y2 + k2*(x)) +
        (x>=x2) * (y3 + k3*(x))
    )

cpdef data_type_t piecewise_five(x, data_type_t x0, data_type_t x1, data_type_t x2, data_type_t x3, data_type_t x4, 
	 								data_type_t y0, data_type_t y1, data_type_t y2, data_type_t y3, data_type_t y4,
	 								data_type_t k0, data_type_t k1, data_type_t k2, data_type_t k3, data_type_t k4):
	
	return (
        (x<x0) * (y0 + k0*(x)) +
        ((x>=x0) & (x<x1)) * (y1 + k1*(x)) +
        ((x>=x1) & (x<x2)) * (y2 + k2*(x)) +
        ((x>=x2) & (x<x3)) * (y3 + k3*(x)) +
        (x>=x3) * (y4 + k4*(x))
    )

cdef class dispersion_model:
	"""
	Dispersion model base class
	"""

	def __init__(self):
		
		# Histograms
		self._h = None
		
		# Emperical fits
		self._p = None
		self._r = None

		self._mu_params = self._r_params = None

		pass

	# Pickling function
	def __reduce__(self):
		x = {}
		x['mu_params'] = self.mu_params
		x['r_params'] = self.r_params
		return (dispersion_model, (), x)

	# Pickling function
	def __setstate__(self, x):
		self.mu_params = x['mu_params']
		self.r_params = x['r_params']

	property h:
		"""Histrogram of observed cleavages at each predicted cleavage rate"""
		def __get__(self):
			return self._h
		def __set__(self, x):
			self._h = x

	property p:
		"""Array of the negative binomial ML fit parameter `p`"""
		def __get__(self):
			return self._p
		def __set__(self, x):
			self._p = x

	property r:
		"""Array of the negative binomial ML fit parameter `r`"""
		def __get__(self):
			return self._r
		def __set__(self, x):
			self._r = x

	property mu_params:
		
		def __get__(self):
			return self._mu_params
		def __set__(self, x):
			self._mu_params = np.array(x, order = 'c')

	property r_params:
		def __get__(self):
			return self._r_params
		def __set__(self, x):
			self._r_params = np.array(x, order = 'c')


	cpdef data_type_t fit_mu(self, data_type_t x):
		"""Computes the mu term for the negative binomial.
		
		:param x: predicted values to be converted to observed means
		
		:returns mu: (float)
		"""

		cdef data_type_t [:] par = self._mu_params
		cdef data_type_t res = piecewise_three(x, *par)

		return res if res > 0.0 else 0.1

	cpdef data_type_t fit_r(self, data_type_t x):
		"""Computes the dispersion term for the negative binomial.
		Note that the model parameters estimate the inverse.

		:param x: predicted values to be converted to dispersion parameter r

		:returns: A fitted paramter for the negative bionomial distribution
		:rtype: float
		"""

		cdef data_type_t [:] par = self._r_params
		cdef data_type_t res = 1.0/piecewise_five(x, *par)

		return res if res > 0.0 else 1e-6

	def __str__(self):
		"""Print model to string"""
		res = "Oops...the functionality has not yet been implemented!"
		return res

	@cython.cdivision(True)
	cpdef data_type_t [:] log_pmf_values(self, data_type_t [:] exp, data_type_t [:] obs):
		"""Computing the probability mass function"""

		cdef int i, n = exp.shape[0]
		cdef double r, mu
		
		cdef data_type_t [:] res = np.zeros(n, dtype = np.float64, order = 'c')
	
		for i in range(n):
			r = self.fit_r(exp[i])
			mu = self.fit_mu(exp[i])
			res[i] = nbinom.logpmf(<int>obs[i], (r/(r+mu)), r)

		return res

	@cython.cdivision(True)
	cpdef data_type_t [:] pmf_values(self, data_type_t [:] exp, data_type_t [:] obs):
		"""Computing the probability mass function"""

		cdef int i, n = exp.shape[0]
		cdef double r, mu
		
		cdef data_type_t [:] res = np.zeros(n, dtype = np.float64, order = 'c')

		for i in range(n):
			r = self.fit_r(exp[i])
			mu = self.fit_mu(exp[i])
			res[i] = nbinom.pmf(<int>obs[i], (r/(r+mu)), r)

		return res


	@cython.cdivision(True)
	cpdef data_type_t [:] pmf_values_0(self, data_type_t [:] exp, data_type_t [:] obs, data_type_t [:] res):
		"""Computing the probability mass function (but writes to a matrix pointer)"""

		cdef int i, n = exp.shape[0]
		cdef double r, mu
		
		#cdef data_type_t [:] res = np.zeros(n, dtype = np.float64, order = 'c')

		for i in range(n):
			r = self.fit_r(exp[i])
			mu = self.fit_mu(exp[i])
			res[i] = nbinom.pmf(<int>obs[i], (r/(r+mu)), r)

		return res

	@cython.cdivision(True)
	cpdef data_type_t [:] log_pmf_values_0(self, data_type_t [:] exp, data_type_t [:] obs, data_type_t [:] res):
		"""Computing the probability mass function (but writes to a matrix pointer)"""

		cdef int i, n = exp.shape[0]
		cdef double r, mu
		
		#cdef data_type_t [:] res = np.zeros(n, dtype = np.float64, order = 'c')

		for i in range(n):
			r = self.fit_r(exp[i])
			mu = self.fit_mu(exp[i])
			res[i] = nbinom.logpmf(<int>obs[i], (r/(r+mu)), r)

		return res

	cpdef data_type_t [:] p_values(self, data_type_t [:] exp, data_type_t [:] obs):
		"""Computes CDF p-value from negative binomial

		:param exp: 
		:param obs:

		:return p:
		"""
		
		cdef int i, n = exp.shape[0]
		cdef double r, mu
		cdef data_type_t [:] res = np.ones(n, dtype = np.float64, order = 'c')

		for i in range(n):
			r = self.fit_r(exp[i])
			mu = self.fit_mu(exp[i])
			res[i] = nbinom.cdf(<int>obs[i], r/(r+mu), r)

		return res

	cpdef resample_p_values(self, data_type_t [:] x, int times):

		cdef int i, j, n = x.shape[0]
		cdef data_type_t r, mu, k

		cdef long [:] vals

		cdef data_type_t [:,:] res_pvals = np.ones((n, times), dtype = np.float64, order = 'c')
		cdef long [:,:] res_vals = np.zeros((n, times), dtype = np.int_, order = 'c')

		for i in range(n):
			r = self.fit_r(x[i])
			mu = self.fit_mu(x[i])
		
			vals = np.random.negative_binomial(r, r/(r+mu), times)
			
			for j in range(times):
				res_pvals[i, j] = nbinom.cdf(<int>vals[j], r/(r+mu), r)
				res_vals[i, j] = vals[j]

		return res_vals, res_pvals


@cython.boundscheck(True)
@cython.wraparound(True)


def learn_dispersion_model(h, cutoff = 250, trim = (2.5, 97.5)):
	"""Learn a dispersion model from the expected vs. observed histogram
	
	:param h: 2-D :class:`numpy.array` containing the distribution of observerd cleavages at each expected cleavage rate
	:type h: numpy.array
	:param cutoff: Mininum number of observed cleavages tp perform ML negative binomial fit
	:type cutoff: int
	:param trim: Percent of data to trim  from the observed cleavage count (to mitigate outlier effects)
	:type trim: tuple of int

	:return: A dispersion model class :class:`dispersion_model`
	:rtype: dispersion_model

	:todo: Add exceptions for failure to fit, etc.
	"""

	size = int(h.shape[0])
	p = np.zeros(size)
	r = np.zeros(size)

	# Make an initial negative binomial fit
	for i in range(0, size):
	
		pos = 0
		x = np.zeros(int(np.sum(h[i,:])))
		for j in range(len(h[i,:])):
			num = int(h[i, j])
			x[pos:pos+num] = j
			pos += num

		# If more than 500k points downsample to
		# make curve-fitting tractable
		if len(x)>1e5:
			x=np.random.choice(x, size=int(1e5))
			x=np.sort(x)

		if len(x) >= cutoff:		

			lower = int(np.floor(x.shape[0]*(trim[0]/100.0)))
			upper = int(np.ceil(x.shape[0]*(trim[1]/100.0)))

			mu = np.mean(x[lower:upper])
			var = np.var(x[lower:upper])

			est_r = (mu * mu) / (var - mu)
			if est_r <= 0.0: 
				est_r = 10.0
			est_p = est_r / (est_r + mu)
			
			(p[i], r[i]) = nbinom.fit(x[lower:upper], p = est_p, r = est_r)

		else:
			p[i] = r[i] = np.nan

	# Back-compute the mean values from the negative binomial parameters
	mus = p*r/(1-p)

	# hack
	r[r>200]=200.

	# Extrapolate using polynomial fit
	x = np.arange(size)
	sele = np.isfinite(mus)

	first_x = np.min(x[sele])
	last_x = np.max(x[sele])*0.75

	# fit mu with a 3 segments
	fit_mu=pwlf.PiecewiseLinFit(x[sele], mus[sele])
	res=fit_mu.fit_with_breaks_force_points(np.linspace(first_x, last_x, 4), [0], [mus[0]])

	# fit r with 5 s segments
	fit_r=pwlf.PiecewiseLinFit(x[sele], 1.0/r[sele])
	res = optimize.minimize(fit_r.fit_with_breaks_opt, [3.0, 7.0, 15.0, 25.0])
	
	x0=np.zeros(6)
	x0[0]=first_x
	x0[-1]=last_x
	x0[1:-1]=res.x

	#res=fit_r.fit_with_breaks_force_points([first_x, 2, 7, 15, 25, last_x], [1], [1.0/r[1]])
	res=fit_r.fit_with_breaks_force_points(x0, [1], [1.0/r[1]])


	# Create a dispersion model class
	res = dispersion_model()
	res.h = h
	res.p = p
	res.r = r

	res.mu_params = list(fit_mu.fit_breaks[1:]) + list(fit_mu.intercepts) + list(fit_mu.slopes)
	res.r_params = list(fit_r.fit_breaks[1:]) + list(fit_r.intercepts) + list(fit_r.slopes)
	
	return res

# Read and write functions for making a portable dispersion model

import json
import base64

def base64encode(x):
	
	return [str(x.dtype), base64.b64encode(x), x.shape]

def base64decode(x):

	dtype = np.dtype(x[0])
	arr = np.frombuffer(base64.b64decode(x[1]), dtype)
	if len(x) > 2:
		return arr.reshape(x[2])
	return arr

def read_dispersion_model(filename):
	
	import json
	import urllib.request as request

	if filename.startswith('http'):
		file = request.urlopen(filename)
	else:
		file = open(filename, 'r')

	params = json.load(file)

	file.close()

	model = dispersion_model()
	model.mu_params = base64decode(params["mu_params"])
	model.r_params = base64decode(params["r_params"])

	if "h" in params:
		model.h = base64decode(params["h"])
	if "p" in params:
		model.p = base64decode(params["p"])
	if "r" in params:
		model.r = base64decode(params["r"])

	return model

def write_dispersion_model(model):
	"""Write a JSON format dispersion model

	:param model: An instance of :class:`dispersion_model`
	:return: JSON-formatted dump of dispersion model
	:rtype: str
	"""

	import json

	out = { "mu_params": base64encode(np.asarray(model.mu_params, order = 'C')), 
			"r_params": base64encode(np.asarray(model.r_params, order = 'C')),
			"h": base64encode(np.asarray(model.h, order = 'C')),
			"p": base64encode(np.asarray(model.p, order = 'C')),
			"r": base64encode(np.asarray(model.r, order = 'C'))
		}

	return json.dumps(out, indent = 4)




# Copyright 2015 Jeff Vierstra

import numpy as np
import scipy.stats

from footprint_tools.stats import windowing

def compute_prior_weighted(fdr, w, cutoff = 0.05, pseudo = 0.5):
	"""Returns prior of whether a TF is bound or not 
	
	Parameters
	----------
	fdr : numpy.array
	    2D :class:`np.array` with FPR values for each datset and position
	w : numpy.array
	    Binary array whether samples is DHS
	cutoff : float, optional
	    FPR cutoff value to label as occupied when building prior
	pseudo : float, optional
	    Psuedocount to add as prior to Beta distribution
	
	Returns
	-------
	TYPE
	    Description
	"""

	k = np.sum(fdr <= cutoff, axis = 0) # num of datasets w/ fp
	n = np.sum(w, axis = 0) #np.sum(w) # num of datasets w/ HS
	A = n-k + pseudo
	B = k + pseudo
	pr = A/(A+B)
	
	res = np.ones(fdr.shape)
	res *= pr[np.newaxis,:]
	res[w==0] = 1

	return res

def compute_delta_prior(obs, exp, fdr, beta_prior, cutoff = 0.05):
	"""Returns a point estimate of exepected cleavage depletion with a footprint per nucleotide 
	
	Parameters
	----------
	obs : TYPE
	    Description
	exp : TYPE
	    Description
	fdr : TYPE
	    Description
	beta_prior : TYPE
	    Description
	cutoff : float, optional
	    Description
	
	Returns
	-------
	TYPE
	    Description
	"""

	(n, w) = obs.shape

	mus = np.ones((n, w))
	ws = np.ones((n, w))
	
	for i in range(n):
		
		k = obs[i,:]
		n = np.max(np.vstack([exp[i,:], obs[i,:]]), axis = 0)

		mu, v = scipy.stats.beta.stats(k+beta_prior[i][0], n-k+beta_prior[i][1], 
			loc = 0, scale = 1, moments = 'mv')

		mus[i,:] = mu
		ws[i,:] = 1/np.sqrt(v)

	ws[fdr > cutoff] = 0

	delta = np.sum(ws*mus, axis = 0) / np.sum(ws, axis = 0)
	delta[np.isnan(delta)] = 1

	return delta

def log_likelihood(obs, exp, dm, delta = 1, w = 3):
	"""
	
	Parameters
	----------
	obs : TYPE
	    Description
	exp : TYPE
	    Description
	dm : TYPE
	    Description
	delta : int, optional
	    Description
	w : int, optional
	    Description
	
	Returns
	-------
	TYPE
	    Description
	"""
	res = np.ones((obs.shape[0], obs.shape[1]), order = 'c')

	n = obs.shape[0]
	for i in range(n):
		res[i,:] = windowing.sum(dm[i].log_pmf_values(exp[i,:] * delta, obs[i,:]), w)
	
	return res

def posterior(prior, ll_on, ll_off):
	"""
	Parameters
	----------
	prior : TYPE
	    Description
	ll_on : TYPE
	    Description
	ll_off : TYPE
	    Description
	
	Returns
	-------
	TYPE
	    Description
	"""
	prior_on = np.log(1-prior)
	prior_off = np.log(prior)

	p_off = prior_off + ll_off
	p_on = prior_on + ll_on
	denom = np.logaddexp(p_on, p_off)

	return (p_off - denom)







# Copyright 2015 Jeff Vierstra
import numpy as np
import itertools
import random


class bias_model(object):

	def __init__(self):

		self.model = {}
		
		self.k = 6
		self.mid = 3

	def __getitem__(self, key):
		
		return self.model.get(key, 1e-6)

	def __setitem__(self, key, value):

		self.model[key] = value

	def offset(self):

		return max(self.k-self.mid, self.mid)

	def shuffle(self):
		"""Randomly shuffle the bias model

		:returns: 
		"""

		ret = bias_model()
		ret.model = { x: y for (x, y) in zip(list(self.model.keys()), sorted(list(self.model.values()), key = lambda k: random.random())) }
		ret.offset = self.offset
		return ret

	def predict(self, probs, n = 100):
		"""Compute cleavage propensities from sequence
		
		:param probs: an array of probilities (relative values)
		:type probs: numpy array (float64)
		:param n: number tags to distrbute
		:type n: integer

		:returns: integer array of relative cleavage counts
		"""

		return np.around( probs / np.sum(probs) * n )

class kmer_model(bias_model):

	def __init__(self, filepath):
		
		bias_model.__init__(self)		
		self.read_model(filepath)
	
	def read_model(self, filepath):
		"""Read the k-mer model from a file.

		:param filepath: the path to a 6-model
		:type filepath: string

		:returns:
		"""

		try:

			for line in open(filepath, 'r'):
				(seq, prob) = line.strip().split('\t')
				self.model[seq.upper()] = float(prob)

		except IOError:
			
			raise IOError("Cannot open file: %s" % filepath)
	
	def probs(self, seq):
		"""Generate cleavage preference array from DNA sequence

		:param seq: DNA sequence

		:returns: array (float)
		"""
		k = self.k
		mid = self.mid
		ofst = self.offset()

		return np.array([ self.__getitem__( seq[(i-mid):(i+(k-mid))] ) for i in range(ofst, len(seq)-ofst) ], dtype = np.float64)

class uniform_model(bias_model):

	def __init__(self):

		bias_model.__init__(self)

		for seq in itertools.product('ATCG', repeat = self.k):
			self.model[''.join(seq)] = 1.0

	def probs(self, seq):
		return np.ones( len(seq) )



# Copyright 2015 Jeff Vierstra

# cython: embedsignature=True


import scipy.special
import scipy.optimize
import numpy as np

cimport cython
cimport numpy as np

cdef extern from "cephes.h":
	double c_exp(double) nogil
	double c_log(double) nogil
	double c_log1p(double) nogil
	double c_lgamma(double) nogil
	double c_incbet(double, double, double) nogil


def mle(par, data, sm):

	"""Objective function for MLE estimate according to
	https://en.wikipedia.org/wiki/Negative_binomial_distribution#Maximum_likelihood_estimation
	
	Parameters
	----------
	par : TYPE
	    Description
	data : TYPE
	    Description
	sm : TYPE
	    Description
	data: the points to be fit
	sm: \sum data / len(data)
	
	Returns
	-------
	TYPE
	    Description
	"""

	p = par[0]
	r = par[1]
	n = len(data)
	f0 = sm / (r+sm) - p
	f1 = np.sum( scipy.special.psi(data+r) ) - n * scipy.special.psi(r) + n * np.log( r / (r+sm) )
	
	return np.array([f0, f1])


def fit(data, p = None, r = None):
	"""Summary
	
	Parameters
	----------
	data : TYPE
	    Description
	p : None, optional
	    Description
	r : None, optional
	    Description
	
	Returns
	-------
	TYPE
	    Description
	"""
	if p is None or r is None:
		av = np.average(data)
		va = np.var(data)
		r = (av*av) / (va-av)
		p = (va-av) / (va)

	sm = np.sum(data)/len(data)
	x = scipy.optimize.fsolve(mle, np.array([p, r]), args = (data, sm))

	return (x[0], x[1])

cpdef data_type_t logpmf(int k, data_type_t p, data_type_t r) nogil:
	cdef double coeff = c_lgamma(k+r) - c_lgamma(k+1) - c_lgamma(r)
	return coeff + r * c_log(p) + k * c_log1p(-p)


cpdef data_type_t pmf(int k, data_type_t p, data_type_t r) nogil:
	return c_exp( logpmf(k, p, r) )

cpdef data_type_t cdf(int k, data_type_t p, data_type_t r) nogil:
	return c_incbet(r, k+1, p)

cpdef data_type_t mean(data_type_t p, data_type_t r) nogil:
	return p*r/(1-p)

"""Not yet implemented
def rvs(data_type_t p, data_type_t r, int n = 1):
	return np.random.negative_binomial(r, p, n)
"""


# Copyright 2015 Jeff Vierstra

# cython: embedsignature=True



cimport numpy as np

ctypedef np.float64_t data_type_t

cpdef segment(data_type_t [:] x, data_type_t threshold, int w = 1, bint decreasing = 0):
	"""Segment an array into continuous elements passing a threshhold

	"""

	cdef double dir = -1 if decreasing else 1

	cdef list ret = []
	cdef int i, curr_start = -1
	for i in range(x.shape[0]):
		if curr_start < 0:
			if dir*x[i] >= dir*threshold:
				curr_start = i-w+1
		else:
			if dir*x[i] < dir*threshold:
				if len(ret) > 0 and curr_start <= ret[-1][1]:
					ret[-1][1] = i-1+w
				else:
					ret.append( [curr_start, i-1+w] )
				curr_start = -1
	return ret



# Copyright 2015 Jeff Vierstra

# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: embedsignature=True


import numpy as np

cimport cython
cimport numpy as np

ctypedef np.float64_t data_type_t

cpdef bisect(data_type_t [:] a, data_type_t [:] b):
	"""
	"""
	cdef int i
	cdef int n = b.shape[0]

	cdef int lo = 0
	cdef int hi = a.shape[0]

	cdef np.ndarray[data_type_t, ndim = 1] ind = np.zeros(n, dtype = np.float64)

	for i in range(n):
		while lo < hi:
			if b[i] < a[lo]: break
			else: lo = lo + 1
		ind[i] = lo

	return ind



# Copyright 2017 Jeff Vierstra

# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: embedsignature=True


import numpy as np
cimport numpy as np

import scipy.stats
import scipy.optimize

ctypedef np.float64_t data_type_t

from ..modeling cimport dispersion

ctypedef dispersion.dispersion_model dm_t

cdef extern from "cephes.h":
	double c_lgamma(double) nogil
	double c_log(double) nogil
	double c_sqrt(double) nogil
	double PI

cpdef np.ndarray[data_type_t, ndim = 3, mode = 'c'] compute_logpmf_values(dm, np.ndarray[data_type_t, ndim = 2, mode = 'c'] obs, np.ndarray[data_type_t, ndim = 2, mode = 'c'] exp,  data_type_t lo, data_type_t hi, int nslices):
	
	cdef int i 		
	cdef int n = obs.shape[0]
	cdef int m = obs.shape[1]

	bins = np.power(2, np.linspace(lo, hi, nslices))

	cdef np.ndarray[data_type_t, ndim = 3, mode = 'c'] e = np.ascontiguousarray(exp[:,:,np.newaxis] * bins)
	cdef np.ndarray[data_type_t, ndim = 3, mode = 'c'] o = np.ascontiguousarray(np.repeat(obs[:,:,np.newaxis], nslices, axis = 2))
	cdef np.ndarray[data_type_t, ndim = 3, mode = 'c'] res = np.zeros((n, m, nslices), order = 'c')

	for i in range(n):
		<dm_t>(dm[i]).log_pmf_values_0(np.ravel(e[i,:,:]), np.ravel(o[i,:,:]), np.ravel(res[i,:,:]))

	return res.T


#various stats functions (should be move to ..stats/distributions folder)

def invchi2_logpdf(x, nu, tau2):
	return (nu/2.0)*np.log(tau2*nu/2.0) + (-nu*tau2/2.0/x) - scipy.special.gammaln(nu/2.0) - (1.0+nu/2.0)*np.log(x)

def invchi2_likelihood(par, x):
	return -np.sum(invchi2_logpdf(x, par[0], par[1]))

cdef double t_logpmf(double x, double nu, double mu, double sig2) nogil:
	return c_lgamma((nu+1.0)/2.0) - c_lgamma(nu/2.0) - c_log(c_sqrt(PI*nu*sig2)) - ((nu+1.0)/2.0 * c_log(1.0+((x-mu)**2)/sig2/nu ))


cpdef np.ndarray[data_type_t, ndim = 2, mode = 'c'] compute_log_prior_t(np.ndarray[data_type_t, ndim = 2, mode = 'c'] obs_over_exp , data_type_t nu_0, data_type_t sig2_0,  data_type_t lo, data_type_t hi, int nslices):
	
	""" 
	Computes the log posterior distribution of the hyperparameters (normal * inv chi2 = students t)
	"""

	cdef int i, j
	cdef int n = obs_over_exp.shape[0]
	cdef int m = obs_over_exp.shape[1]

	cdef data_type_t mu, ssqdev, nu_1, sig2_1

	cdef np.ndarray[data_type_t, ndim = 2, mode = 'c'] res = np.zeros((m, nslices), dtype = np.float64, order = 'c')
	cdef data_type_t [:,:] res_view = res

	cdef double h = (hi-lo)/<double>nslices

	for i in range(m):
		
		mu = np.nanmean(obs_over_exp[:,i])
		ssqdev = np.nansum((obs_over_exp[:,i]-mu)**2)

		nu_1 = nu_0 + n
		sig2_1 = (nu_0*sig2_0 + ssqdev) / (nu_0 + n)

		with nogil:

			for j in range(0, nslices):
				#res_view[i, j] = scipy.stats.t.logpdf(bins[j], df = nu_1, loc = mu, scale = np.sqrt(sig2_1))
				res_view[i, j] = t_logpmf(lo+(j*h), nu_1, mu, sig2_1)

		res[i, :] += np.log(h)

	return res.T # - res.sum(axis = 1)[:, np.newaxis]).T







# Copyright 2015 Jeff Vierstra

# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: embedsignature=True


from libc.stdlib cimport free


import numpy as np

cimport numpy as np

ctypedef np.float64_t data_type_t

ctypedef double (*func_t)(const double *, int)
ctypedef double (*weighted_func_t)(const double* const, const double* const, int)

cdef extern from "windowing.h":
    #
    double fast_sum(const double* const, int)
    double fast_product(const double* const, int)
    double fast_fishers_combined(const double* const, int)
    double fast_stouffers_z(const double* const, int)
    double* fast_windowing_func(const double* const, int, int, func_t)
    #
    double fast_weighted_stouffers_z(const double* const, const double* const, int)
    double* fast_weighted_windowing_func(const double* const, const double* const, int, int, weighted_func_t)

cdef windowing_func(data_type_t [:] x, int hw, func_t func_ptr):
    
    cdef int i
    cdef int n = x.shape[0]

    cdef double* res = fast_windowing_func(&x[0], n, hw, func_ptr)
    cdef data_type_t [:] win = np.ones(n, dtype = np.float64, order = 'c')
    
    for i in range(hw, n-hw):
        win[i] = res[i]
    
    free(res)

    return np.asarray(win)

cpdef sum(data_type_t [:] x, int hw):
    """Sum of values in sliding local window

    :param x: (array) values
    :param hw: (int) half window size to smooth
    
    :return: (float) summed values
    """

    return windowing_func(x, hw, fast_sum)

cpdef product(data_type_t [:] x, int hw):
    """Product of values in sliding local window

    :param x: (array) values
    :param hw: (int) half window size to smooth
    
    :return: (float) product of values
    """
    
    return windowing_func(x, hw, fast_product)

cpdef fishers_combined(data_type_t [:] x, int hw):
    """Compute p-value for a window using Fisher's combined method

    :param x: (array) p-values
    :param hw: (int) half window size to smooth
    
    :return p: (float) combined p-values
    """


    return windowing_func(x, hw, fast_fishers_combined)

cpdef stouffers_z(data_type_t [:] x, int hw):
    """Compute p-value for a window using Stouffer's method

    :param x: (array) p-values
    :param hw: (int) half window size to smooth
    
    :return p: (float) combined p-values
    """

    return windowing_func(x, hw, fast_stouffers_z)

cdef weighted_windowing_func(data_type_t [:] x, data_type_t [:] w, int hw, weighted_func_t func_ptr):
    
    cdef int i
    cdef int n = x.shape[0]

    cdef double* res = fast_weighted_windowing_func(&x[0], &w[0], n, hw, func_ptr)
    cdef data_type_t [:] win = np.ones(n, dtype = np.float64, order = 'c')
    
    for i in range(hw, n-hw):
        win[i] = res[i]
    
    free(res)

    return np.asarray(win)

cpdef weighted_stouffers_z(data_type_t [:] x, data_type_t [:] w, int hw):
    """Compute p-value for a window using weighted Stouffer's method

    :param x: (array) p-values
    :param w: (array) weights
    :param hw: (int) half window size to smooth
    
    :return p: (float) combined p-values
    """

    return weighted_windowing_func(x, w, hw, fast_weighted_stouffers_z)



# Copyright (C) 2015-2020 Jeff Vierstra (jvierstra@altius.org)


import os
import sys
import pathlib


from setuptools import find_packages, setup
from distutils.command.build_clib import build_clib
from distutils.extension import Extension
from distutils import log
from Cython.Distutils import build_ext

import numpy as np

from glob import glob

__version__ = "1.1.4"

if sys.version_info[0] != 3 or sys.version_info[1] <= 5:
    print("Package requires Python version 3.5+")
    sys.exit(1)

cephes_include = "cephes"
cephes_src = glob("cephes/*.c")
cehpes_lib = ('cephes', { 'sources': cephes_src })
# see MANIFEST.in -- a slight hack to include all of the header files in the source distrubution

modules = [
	dict(name="footprint_tools.modeling.predict", sources=["footprint_tools/modeling/predict.pyx"]),
	dict(name="footprint_tools.modeling.predict", sources=["footprint_tools/modeling/predict.pyx"]),
	dict(name="footprint_tools.modeling.dispersion", sources=["footprint_tools/modeling/dispersion.pyx"]),
	dict(name="footprint_tools.stats.windowing", sources = ["footprint_tools/stats/windowing.pyx"]),
	dict(name="footprint_tools.stats.distributions.nbinom", sources=["footprint_tools/stats/distributions/nbinom.pyx"]),
	dict(name="footprint_tools.stats.segment", sources=["footprint_tools/stats/segment.pyx"]),
	dict(name="footprint_tools.stats.fdr.bisect", sources=["footprint_tools/stats/fdr/bisect.pyx"]),
	dict(name="footprint_tools.stats.differential", sources=["footprint_tools/stats/differential.pyx"])
]

scripts = ["scripts/ftd-learn-dispersion-model", "scripts/ftd-compute-deviation", "scripts/ftd-compute-posterior", "scripts/ftd-learn-beta-prior", "scripts/ftd-diff-test"]
install_requires = ["cython", "numpy>=1.10", "scipy>=0.17", "pysam>=0.15", "pyfaidx>=0.4.2", "statsmodels", "genome_tools>=1.0.2", "pwlf"]

setup(
	name = "footprint_tools",
	version = __version__,
	license = "GPL-3.0-or-later",
	description = "Genomic footprint detection",
	long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
	author = "Jeff Vierstra",
	author_email = "jvierstra@altius.org",
	url = "https://github.com/jvierstra/footprint-tools",
	download_url = "https://github.com/jvierstra/footprint-tools/archive/1.1.3.tar.gz",
	keywords = ["genomic footprints", "bioinformatics"],
	zip_safe = False,
	packages =  find_packages(),
	libraries = [cehpes_lib],
    ext_modules = [Extension(**opts) for opts in modules],
    include_dirs=[np.get_include(), cephes_include],
    cmdclass = {'build_clib': build_clib, 'build_ext': build_ext},
    install_requires = install_requires,
    scripts = scripts,
    classifiers=[
	    'Development Status :: 5 - Production/Stable', 
	    'Intended Audience :: Science/Research', 
	    'Topic :: Scientific/Engineering :: Bio-Informatics',
	    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
	    'Programming Language :: Python :: 2.7',
	    'Programming Language :: Python :: 3',
],
)



# Copyright 2015 Jeff Vierstra

import genome_tools
import pysam

import numpy as np
from collections import defaultdict

class ReadError(Exception):
	pass

class GenotypeError(Exception):
	pass

class bamfile(object):
	"""Class to access BAM files
	
	Attributes
	----------
	min_qual : int
	    Filter reads by minimim mapping quality (MAPQ)
	offset : tuple
	    Position offsets to apply to the `+` and `-` strands (default =(0, -1))
	remove_dups : bool
	    Remove reads with duplicate flag (512) set
	remove_qcfail : bool
	    Remove reads with QC fail flag (1024) set
	samfile : pysam.Samfile
	    SAM/BAM file object
	"""
	
	def __init__(self, filepath, min_qual = 1, remove_dups = False, remove_qcfail = True, offset = (0, -1)):
		"""Constructor
		
		Parameters
		----------
		filepath : str
		    Fileath to SAM/BAM file
		min_qual : int, optional
		    Filter reads by minimim mapping quality (MAPQ)
		remove_dups : bool, optional
		    Remove reads with duplicate flag (512) set
		remove_qcfail : bool, optional
		    Remove reads with QC fail flag (1024) set
		offset : tuple, optional
		    Position offsets to apply to the `+` and `-` strands (default =(0, -1))
		
		Raises
		------
		IOError
		    If file is not found, has no index, or something else bad happened while trying to open BAM file
		"""

		try:
			self.samfile = pysam.Samfile(filepath, "rb")
		except:
			raise IOError("Cannot open BAM file: %s" % filepath)

		self.offset = offset #-1 # a hack for the mis-aligned data from 2010
		self.min_qual = min_qual
		self.remove_dups = remove_dups
		self.remove_qcfail = remove_qcfail

	def validate_read(self, read):
		"""
		Validate BAM tag
		
		Parameters
		----------
		read : pysam.AlignedSegment
		    Read from BAM/SAM file
		
		Returns
		-------
		pysam.AlignedSegment
		    Same read as input
		
		Raises
		------
		ReadError
		    Raises error if read fails QC flag, is a duplicate or MAPQ < minimum
		"""

		if self.remove_qcfail and read.is_qcfail:
			raise ReadError()
		if self.remove_dups and read.is_duplicate:
			raise ReadError()
		if read.mapping_quality < self.min_qual:
			raise ReadError()

		return read

	def get_read_mate(self, read):
		"""
		Fetch the mate pair for paired-end reads
		
		Parameters
		----------
		read : pysam.AlignedSegment
			One  of the read-pairs
		
		Returns
		-------
		pysam.AlignedSegment
		    The corresponding mate read
		
		"""

		fpos = self.samfile.tell()
		try:
			mate = self.samfile.mate(read)
		except ValueError:
			mate = None
		finally:
			self.samfile.seek(fpos)
		return mate

	def read_pair_generator(self, chrom, start, end):
		"""
		Generator function that returns sequencing tags within a given region
		
		Parameters
		----------
		chrom : str
		    Chromosome
		start : int
		    Start coordinate
		end : int
		    End coordinate
		
		Yields
		------
		tuple
		    A tuple of :class:`pysam.AlignedSegments`. Elements may be `NoneType` if single-end sequencing or one of pairs falls outisde of query range.
		
		"""

		read_dict = defaultdict(lambda: [None, None])

		for read in self.samfile.fetch(chrom, max(start-10, 0), end+10):

			try:

				self.validate_read(read)

				# Single-end:,just pass the read
				if not read.is_paired:
					yield read, None 
				
				#  Pair-end; do some validation
				else:

					if not read.is_proper_pair or read.is_secondary or read.is_supplementary:
						continue

					qname = read.query_name

					if qname not in read_dict:
						if read.is_read1:
							read_dict[qname][0] = read
						else:
							read_dict[qname][1] = read
					else:
						if read.is_read1:
							yield read, read_dict[qname][1]
						else:
							yield read_dict[qname][0], read
						del read_dict[qname]

			except ReadError as e:
				continue

		"""Flush out the rest of dictionary. (for example if one the 
		mates wasn't in the original region). It might be reasonable to 
		manually grab the remaining read using `__get_read_mate`"""
		for k, reads in read_dict.items():
			yield reads[0], reads[1]


	def __add_read(self, read, fw, rev):
		"""

		"""
		if read.is_reverse:
			a = int(read.reference_end)+self.offset[1]
			rev[a] = rev.get(a, 0.0) + 1.0
		else:
			a = int(read.reference_start)+self.offset[0]
			fw[a] = fw.get(a, 0.0) + 1.0

	def __lookup(self, interval):
		"""

		"""	
		chrom = interval.chrom
		start = interval.start
		end = interval.end
		flip = True if interval.strand == '-' else False

		tmp_fw = {}
		tmp_rev = {}

		for read1, read2 in self.read_pair_generator(chrom, max(start-10, 0), end+10):
			if read1:
				self.__add_read(read1, tmp_fw, tmp_rev)
			if read2:
				self.__add_read(read2, tmp_fw, tmp_rev)

		fw_cutarray = np.array([tmp_fw.get(i, 0.0) for i in range(start, end)])
		rev_cutarray = np.array([tmp_rev.get(i, 0.0) for i in range(start, end)])

		return {
 			"+": rev_cutarray[::-1] if flip else fw_cutarray, 
			"-": fw_cutarray[::-1] if flip else rev_cutarray
			}

	def __validate_genotype(self, read, pos, ref, alt):
		"""

		Returns
		-------
		bool
			True if read contains reference allele and passes filters
			False if read contains alternate allele and passes filters

		Raises
		------
		IndexError
		"""
		if not read:
			return 0

		var_offset=pos-read.reference_start

		try:
			base_call=read.query_sequence[var_offset]
		except IndexError as e:
			return 0

		mismatches=int(read.get_tag("XM", with_value_type=False))
		if base_call==ref and mismatches<=1:
			return ref
		if  base_call==alt and mismatches<=2:
			return alt
			raise GenotypeError()



		return read.flag & (1<<12)

	def __lookup_allelic(self, chrom, start, end, pos, ref, alt, flip=False):
		"""

		"""
		#visited_read_pairs = set()

		tmp_ref_fw = {}
		tmp_ref_rev = {}
		tmp_alt_fw = {}
		tmp_alt_rev = {}

		for read1, read2 in self.read_pair_generator(chrom, max(start-10, 0), end+10):

			#try read1 for genotype
			try:
				# Check reads for proper genotypes; raise Genotype error
				# if some thing is problematic, if a read pair doesn't exist or
				# doesn't overlap the SNV then returns 0
				read1_gt=self.__validate_genotype(read1)
				read2_gt=self.__validate_genotype(read2)

				if (not read1_gt) and (not read2_gt): # neither read overlaps SNV
					continue # got to get read pair
				elif (read1_gt and read2_gt) and read1_gt!=read2_gt: # discordant genotypes
					raise GenotypeError("Discordant genotypes between mate reads!")
				elif read1_gt==ref or read2_gt==ref: # Matches REF allele
					fw=tmp_ref_fw
					rev=tmp_ref_rev
				elif read1_gt==alt or read2_gt==alt: # Matches ALT allele
					fw=tmp_alt_fw
					rev=tmp_alt_rev
				else:
					raise GenotypeError()

				if read1:
					self.__add_read(read1, fw, rev)
				if read2:
					self.__add_read(read2, fw, rev)				

			except GenotypeError as e:
				continue
	
		ref_fw_cutarray = np.array([tmp_ref_fw.get(i, 0.0) for i in range(start, end)])
		ref_rev_cutarray = np.array([tmp_ref_rev.get(i, 0.0) for i in range(start, end)])
		alt_fw_cutarray = np.array([tmp_alt_fw.get(i, 0.0) for i in range(start, end)])
		alt_rev_cutarray = np.array([tmp_alt_rev.get(i, 0.0) for i in range(start, end)])

		return {
			ref: {
				"+": ref_rev_cutarray[::-1] if flip else ref_fw_cutarray, 
				"-": ref_fw_cutarray[::-1] if flip else ref_rev_cutarray
				},
			alt: {
				"+": alt_rev_cutarray[::-1] if flip else alt_fw_cutarray, 
				"-": alt_fw_cutarray[::-1] if flip else alt_rev_cutarray
			}
		}


	def __getitem__(self, x):
		"""Getter function to retrieve cutcounts
		
		Parameters
		----------
		x : genome_tools.genomic_interval or pysam.VariantRecord
		    Retrieve cleavages over a windowed region or resolve allelically over a known variant
		
		Returns
		-------
		dict
		    A dictionary of read counts resolved to each strand, and each allele (if approriate)
		
		Raises
		------
		TypeError
		    If input is neither a genome_tools.genomic_interval nor a pysam.VariantRecord
		"""
		if isinstance(x, genome_tools.genomic_interval):
			return self.__lookup(x)
		elif isinstance(x, pysam.VariantRecord):
			return self.__lookup_allelic(x)
		else:
			raise TypeError()



# Copyright 2015 Jeff Vierstra

# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: embedsignature=True

cimport numpy as np

import numpy as np

cdef extern from "predict.h":
	struct result:
		double* exp
		double* win	
	void free_result_t(result*)
	result* fast_predict(double*, double*, int, int, int, double)

ctypedef np.float64_t data_type_t

cdef predict(data_type_t [:] obs, data_type_t [:] probs, int half_window_width, int smoothing_half_window_width, double smoothing_clip):

	cdef int i
	cdef int l = obs.shape[0]

	# Predict
	cdef result* res = fast_predict(&obs[0], &probs[0], l, half_window_width, smoothing_half_window_width, smoothing_clip)

	# Copy back into a python object
	cdef np.ndarray[data_type_t, ndim = 1, mode = 'c'] exp = np.zeros(l, dtype = np.float64, order = 'c')
	cdef data_type_t [:] exp_view = exp

	cdef np.ndarray[data_type_t, ndim = 1, mode = 'c'] win = np.zeros(l, dtype = np.float64, order = 'c')
	cdef data_type_t [:] win_view = win

	for i in range(l): 
		exp_view[i] = res.exp[i]
		win_view[i] = res.win[i]

	# Free up the memory from C
	free_result_t(res)

	return exp, win

def reverse_complement(seq):
	"""Computes reverse complement of a DNA sequence
	
	Parameters
	----------
	seq : str
	    DNA sequence string
 	
	Returns
	-------
	str
	    Reverse complement of ``seq``
	"""
	compl = { 'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'n': 'n'}
	return ''.join([ compl.get(base, 'N') for base in seq ])[::-1]

class prediction(object):

	"""Class that holds a wrapper function to 
		compute the expected cleavage counts
	
	Attributes
	----------
	bm : bias.bias_model
	    Bias model class
	counts : dict
	    Dictionary of observed cleavage counts {'+': [...], '-': [...]}
	half_window_width : int
	    Window width to apply bias model (final windows size = 2W+1)
	interval : genome_tools.genomic_interval
	    Genomic region to computed expected cleavages
	padding : int
	    Padding applied to ``interval`` when retrieving per-nucleotide data
	seq : str
	    FASTA sequence within 
	smoothing_clip : float
	    Fraction of nucleotides to trim when computing smoothed mean
	smoothing_half_window_width : int
	    Desc
	"""
	
	def __init__(self, reads, fasta, interval, bm, half_window_width = 5, smoothing_half_window_width = 0, smoothing_clip = 0.0):
		
		self.bm = bm
		self.half_window_width = half_window_width
		self.smoothing_half_window_width = smoothing_half_window_width
		self.smoothing_clip = smoothing_clip

		self.padding = self.half_window_width + smoothing_half_window_width

		self.interval = interval

		pad_interval = interval.widen(self.padding)
		
		# Note: We clip the first base when recombining the positive 
		# and negative strand, so add an extra base upfront
		pad_interval.start -= 1

		self.counts = reads[pad_interval]
		self.seq = fasta[pad_interval.chrom][pad_interval.start-bm.offset():pad_interval.end+bm.offset()].seq.upper()

	def compute(self):
		"""Computes the expected cleavage counts from observed 
			counts, a FASTA sequence and a sequence bias model
		
		Returns
		-------
		tuple
		    (observerd, expected, wimdowed per-nucleotide counts)
		"""
		obs_counts = {'+': None, '-': None}
		exp_counts = {'+': None, '-': None}
		win_counts = {'+': None, '-': None}

		for strand in ['+', '-']:

			# Pre-calculate the sequence bias propensity table from bias model
			if strand == '+':
				probs = self.bm.probs(self.seq)
			else:
				probs = self.bm.probs(reverse_complement(self.seq))[::-1]

			exp, win = predict(np.ascontiguousarray(self.counts[strand]), np.ascontiguousarray(probs), self.half_window_width, self.smoothing_half_window_width, self.smoothing_clip)

			w = self.counts[strand].shape[0] - self.padding

			obs_counts[strand] = self.counts[strand][self.padding:w]
			exp_counts[strand] = exp[self.padding:w]
			win_counts[strand] = win[self.padding:w]

		return (obs_counts, exp_counts, win_counts)

